/**
 * @file drvAsynUSBHID.cpp
 * @brief  USB HID 用 AsynPortDriver
 * @author Tetsuya Michikawa(tounichi)
 * @date 2017/7/4
 * @detail
 * USB HIDデバイスとの通信を行う。
 * 実際のUSB HIDデバイスとの通信には HIDAPI を使用するので、
 * OS側で、HIDAPIライブラリをインストールする必要がある。
 */
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cerrno>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <unistd.h>

#include <epicsEvent.h>
#include <epicsExport.h>
#include <epicsMutex.h>
#include <epicsString.h>
#include <epicsThread.h>
#include <epicsTime.h>
#include <epicsTypes.h>
#include <iocsh.h>

#include "asynUSBHID.h"
#include "drvAsynUSBHID.h"

static asynUSBHID usbhidDummy = {};
epicsShareDef asynUSBHID *pasynUSBHID = &usbhidDummy;

/**
 * @brief クラスコンストラクタ
 *	  asynPortDeriverを継承
 * @param[in]	portName	asynで管理するための接続名
 * @param[in]	vendorID	USBデバイスの vendor ID
 * @param[in] 	productID	USBデバイスの product ID
 * @param[in]	serialNo	USBデバイスのシリアル番号(同一デバイスを複数接続する場合に設定する。無い場合にはNULLを設定)
 * @param[in]	autoConnect	自動接続フラグ(0:非自動接続,1:自動接続)
 * @detail 
 */
drvAsynUSBHID::drvAsynUSBHID(const char *portName,
			     const WORD vendorID,
			     const WORD productID,
			     const char *serialNo,
			     int autoConnect ):
  asynPortDriver( portName,
                  0, // maxAddr
                  0, // paramTableSize
                  asynCommonMask | asynOctetMask | asynDrvUserMask, // Interface mask
                  asynCommonMask | asynOctetMask,  // Interrupt mask
                  ASYN_CANBLOCK, // asynFlags
                  autoConnect,  // Autoconnect
                  0,  // Default priority
                  0 ) // Default stack size
{
  asynStatus status;

  // USBデバイス情報を設定
  this->serial_no = NULL;

  if(serialNo != NULL) {
    if(strlen(serialNo) > 0) {
      this->serial_no = epicsStrDup(serialNo);
    }
  }
  this->vendor_id  = vendorID;
  this->product_id = productID;

  // デバイス名を作成
  memset(deviceNameStr, 0x00, MAX_STR);
  if(this->serial_no != NULL) {
    sprintf(deviceNameStr, "%X:%X:%s", this->vendor_id, this->product_id, this->serial_no);
  } else {
    sprintf(deviceNameStr, "%X:%X", this->vendor_id, this->product_id);
  }

  this->usbhid.interfaceType = asynUSBHIDType;
  this->usbhid.pinterface    = pasynUSBHID;
  this->usbhid.drvPvt        = this;

  status = pasynManager->registerInterface( portName, &this->usbhid );
  if( status ) {
    asynPrint( pasynUserSelf, ASYN_TRACE_ERROR,
               "drvAsynUSBHID::drvAsynUSBHID, error registering USBHID interface\n" );
  }
  if(autoConnect) {
    pasynManager->exceptionDisconnect( pasynUserSelf );
    status = this->connect( pasynUserSelf );
    if( status ) {
      asynPrint( pasynUserSelf, ASYN_TRACE_ERROR,
                 "drvAsynUSBHID::drvAsynUSBHID, error calling connect %s\n",
                 pasynUserSelf->errorMessage );
    }
  }
}


/**
 * @brief
 *  
 * @param[in]	pasynUser	asyn識別子
 * @return asynSuccess
 */
asynStatus
drvAsynUSBHID::flushOctet( asynUser *pasynUser )
{
  return asynSuccess;
}

/**
 * @brief データ受信処理
 * 	8bit単位でのデータ受信を行う
 * @param[in] 	pasynUser	asyn識別子
 * @param[out]	value		受信データ
 * @param[in]	maxChars	最大受信データ長(?)
 * @param[out]	nActual		受信データ長
 * @param[out]	eomReason	
 * @retval	asynSuccess	成功
 * @retval	asynError	失敗
 */
asynStatus
drvAsynUSBHID::readOctet( asynUser *pasynUser,
			  char *value,
			  size_t maxChars,
			  size_t *nActual,
			  int *eomReason )
{
  if(this->hid_Hndl == NULL) {
    epicsSnprintf(pasynUser->errorMessage,pasynUser->errorMessageSize,
                  "%s: %s not open!", portName, this->deviceNameStr);
    return asynError;
  }

  BYTE readBuf[MAXCHAR];
  int  rtn;
  asynStatus status = asynSuccess;
  if( eomReason ) {
    *eomReason = 0;
  }

  int tmo_ms = (int)( pasynUser->timeout * 1.e3 );
  if(tmo_ms <= 0) {
    rtn = hid_read(this->hid_Hndl, readBuf, MAXCHAR);
  } else {
    rtn = hid_read_timeout(this->hid_Hndl, readBuf, MAXCHAR, tmo_ms);
  }
  *nActual = rtn;
  memcpy(value, readBuf, rtn);

  asynPrint( pasynUser, ASYN_TRACEIO_DRIVER,
             "%s: read %lu bytes from %s, return %s\n",
             portName, (unsigned long)*nActual, this->deviceNameStr,
             pasynManager->strStatus( status ) );

  asynPrintIO(pasynUser, ASYN_TRACEIO_DRIVER, (char *)readBuf, rtn,
	      "%s read(raw) %lu\n", portName, (unsigned long)rtn);

  return status;
}

/**
 * @brief データ送信処理
 *	8bit単位でのデータ送信を行う
 * @param[in] 	pasynUser	asyn識別子
 * @param[in]	value		送信データ
 * @param[in]	maxChars	最大送信データ長
 * @param[out]	nActual		送信データ長
 * @retval	asynSuccess	成功
 * @retval	asynError	失敗
 */
asynStatus
drvAsynUSBHID::writeOctet( asynUser *pasynUser,
			   char const *value,
			   size_t maxChars,
			   size_t *nActual )
{
  if(this->hid_Hndl == NULL) {
    epicsSnprintf(pasynUser->errorMessage,pasynUser->errorMessageSize,
                  "%s: %s not open!", portName, this->deviceNameStr);
    return asynError;
  }

  if(maxChars == 0) {
    *nActual = 0;
    return asynSuccess;
  }

  int rtn;
  char wrtBuf[MAXCHAR];
  asynStatus status = asynSuccess;

  memset(wrtBuf, 0x00, MAXCHAR);
  if(maxChars <= MAXCHAR) {
    memcpy(wrtBuf, value, maxChars);
  } else {
    epicsSnprintf(pasynUser->errorMessage,pasynUser->errorMessageSize,
                  "%s: %s buffer size > 65", portName, this->deviceNameStr);
    return asynError;
  }

  asynPrintIO(pasynUser, ASYN_TRACEIO_DRIVER, wrtBuf, maxChars,
	      "%s write %lu\n", portName, (unsigned long)maxChars);

  asynPrintIO(pasynUser, ASYN_TRACEIO_DRIVER, wrtBuf, MAXCHAR,
	      "%s write(raw) %lu\n", portName, (unsigned long)MAXCHAR);

  
  rtn = hid_write(this->hid_Hndl, (unsigned char *)wrtBuf, MAXCHAR);
  if(rtn >= 0) {
    *nActual = maxChars;
  }    

  asynPrint( pasynUser, ASYN_TRACEIO_DRIVER,
             "%s: wrote %lu bytes to %s, return %s\n",
             portName, (unsigned long)*nActual, this->deviceNameStr,
             pasynManager->strStatus( status ) );

  return status;
}

/**
 * @brief 接続処理
 *	USBデバイスへ接続する
 * @param[in] 	pasynUser	asyn識別子
 * @retval	asynSuccess	成功
 * @retval	asynError	失敗
 * @detail USBデバイスへの接続は、non blocking のみとする
 */
asynStatus
drvAsynUSBHID::connect( asynUser *pasynUser )
{
  if(this->hid_Hndl != NULL) {
    epicsSnprintf(pasynUser->errorMessage,pasynUser->errorMessageSize,
		  "%s: %s is already open!", portName, this->deviceNameStr);
    return asynError;
  }

  asynPrint( pasynUser, ASYN_TRACEIO_DRIVER,
             "%s: Open connection to %s\n", portName, this->deviceNameStr );
  wchar_t *pWsn = NULL;
  wchar_t wsn[MAX_STR];
  memset(wsn, 0x00, sizeof(wsn));
  if(this->serial_no != NULL) {
    mbstowcs(wsn, this->serial_no, strlen(this->serial_no));
    pWsn = wsn;
  }
  
  // hid device open
  this->hid_Hndl = hid_open(this->vendor_id, this->product_id, pWsn);
  if(this->hid_Hndl == NULL) {
    epicsSnprintf( pasynUser->errorMessage,pasynUser->errorMessageSize,
                   "%s: Can't open %s", this->portName, this->deviceNameStr);
    return asynError;
  }

  // set non blocking mode
  if(hid_set_nonblocking(this->hid_Hndl, 1) < 0) {
    epicsSnprintf( pasynUser->errorMessage,pasynUser->errorMessageSize,
		   "%s: Can't set non blocking mode %s", this->portName, this->deviceNameStr);
    return asynError;
  }

  pasynManager->exceptionConnect( pasynUser );
  return asynSuccess;
}

/**
 * @brief 切断処理
 *	USBデバイスを切断
 * @param[in] 	pasynUser	asyn識別子
 * @retval	asynSuccess	成功
 * @retval	asynError	失敗
 */
asynStatus
drvAsynUSBHID::disconnect( asynUser *pasynUser )
{
  if(this->hid_Hndl == NULL) {
    epicsSnprintf(pasynUser->errorMessage,pasynUser->errorMessageSize,
                  "%s: %s not open!", portName, this->deviceNameStr);
    return asynError;
  }
  hid_close(this->hid_Hndl);
  this->hid_Hndl = NULL;

  pasynManager->exceptionDisconnect(pasynUser);
  return asynSuccess;
}

extern "C" {
  int drvAsynUSBHIDConfigure(const char *portName,
			     const WORD vendorID,
			     const WORD productID,
			     const char *serailNo,
			     int autoConnect = 1 ) {
    new drvAsynUSBHID(portName, vendorID, productID, serailNo, autoConnect );
    return( asynSuccess );
  }
  static const iocshArg initUSBHIDArg0 = { "portName", iocshArgString };
  static const iocshArg initUSBHIDArg1 = { "vendorID", iocshArgInt };
  static const iocshArg initUSBHIDArg2 = { "productID",  iocshArgInt };
  static const iocshArg initUSBHIDArg3 = { "serialNumber", iocshArgString };
  static const iocshArg initUSBHIDArg4 = { "autoConnect",  iocshArgInt };
  static const iocshArg * const initUSBHIDArgs[] = { &initUSBHIDArg0, &initUSBHIDArg1, &initUSBHIDArg2, &initUSBHIDArg3, &initUSBHIDArg4 };
  static const iocshFuncDef initUSBHIDFuncDef = { "drvAsynUSBHIDConfigure", 5, initUSBHIDArgs };
  static void initUSBHIDCallFunc( const iocshArgBuf *args ) {
    drvAsynUSBHIDConfigure( args[0].sval, args[1].ival, args[2].ival, args[3].sval, args[4].ival );
  }

  void drvAsynUSBHIDRegister( void ) {
    static int firstTime = 1;
    if ( firstTime ) {
      iocshRegister( &initUSBHIDFuncDef, initUSBHIDCallFunc );
      firstTime = 0;
    }
  }

  epicsExportRegistrar( drvAsynUSBHIDRegister );
}
